#ifndef RNP2TO191_CK
#define RNP2TO191_CK
//
//			Yasuharu Okamoto, 2012.05, 2013.03
//
/************************************************

      R. L. Eubank and A. Kupresanin
	  "Statistical Computing in C++ and R", 2012, P. 129.

	  The period is approximately 2**191 (about 3 * 10**57)
	  if
	    x_1_0, x_1_1, x_1_2 are not all 0,
	    and all are less than 2**32 - 209,
      and 
	    x_2_0, x_2_1, x_2_1 are not all 0,
	    and all are less than 2**32 - 22853.

************************************************/

using namespace System;

namespace rnp2to191 {

	class RNP2to191 {
		public:
				__int64 x1_0, x1_1, x1_2, x1_3;
				__int64 x2_0, x2_1, x2_2, x2_3;
				__int64 c2p32minus209, c2p32minus22853;
				double inv2p32minus208, cforzeroZn;

				inline __int64 DivModP( __int64 a, __int64 m ){
						__int64 v = a % m;
						if (v < 0) v += m;
						return v;
				}

				inline void SetCoeff(){
					c2p32minus209 = __int64(0x0000000100000000) - 209;
					c2p32minus22853 = __int64(0x0000000100000000) - 22853;
					inv2p32minus208 = 1.0 / (Math::Pow(2.0, 32.0) - 208.0);
					cforzeroZn = (Math::Pow(2.0, 32.0) - 209.0) / (Math::Pow(2.0, 32.0) - 208.0);

					return;
				}

				RNP2to191(){
					SetCoeff();
					
					x1_2 = 1;
					x1_1 = 2;
					x1_0 = 3;
					x2_2 = 4;
					x2_1 = 5;
					x2_0 = 6;
				}

				RNP2to191( __int64 a1, __int64 a2, __int64 a3, __int64 a4, __int64 a5, __int64 a6 ){
					SetCoeff();
					
					x1_2 = a1;
					x1_1 = a2;
					x1_0 = a3;
					x2_2 = a4;
					x2_1 = a5;
					x2_0 = a6;
				}

				void Init(){
					SetCoeff();
					
					x1_2 = 1;
					x1_1 = 2;
					x1_0 = 3;
					x2_2 = 4;
					x2_1 = 5;
					x2_0 = 6;
				}

				void Init( __int64 a1, __int64 a2, __int64 a3, __int64 a4, __int64 a5, __int64 a6 ){
					SetCoeff();
					
					x1_2 = a1;
					x1_1 = a2;
					x1_0 = a3;
					x2_2 = a4;
					x2_1 = a5;
					x2_0 = a6;
				}

				//
				//			0 < uni() < 1
				//
				double uni(){
					x1_3 = x1_2; 
					x1_2 = x1_1;
					x1_1 = x1_0;
					x2_3 = x2_2;
					x2_2 = x2_1;
					x2_1 = x2_0;
					x1_0 = DivModP( 1403580 * x1_2 - 810728 * x1_3, c2p32minus209 );
					x2_0 = DivModP( 527612 * x2_1 - 1370589 * x2_3, c2p32minus22853 );

					__int64 zn = DivModP( x1_0 - x2_0, c2p32minus209 );
					if (zn > 0){
							return zn * inv2p32minus208;
					}else{
							return cforzeroZn;
					}
				}

				inline __int64 MultDivModP( __int64 A, __int64 B, __int64 M ){
					__int64 p2to31 = __int64(0x0000000080000000L);
					__int64 ah, bh, a, b;
					ah = A / p2to31;
					a = A % p2to31;
					bh = B / p2to31;
					b = B % p2to31;
					__int64 m0 = (__int64(0x4000000000000000L) * ah * bh) % M;
					__int64 a0 = DivModP( bh * p2to31 * a, M );
					__int64 b0 = DivModP( ah * p2to31 * b, M );
					__int64 c0 = DivModP( a * b, M );
					__int64 R = DivModP( m0 + a0 + b0 + c0, M );
					return R;
				}

				void Jump2toj( int j ){
					if (j < 1) return;

					__int64 a[3][3], b[3][3];
					a[0][0] = 0;		a[0][1] = 1;	a[0][2] = 0;
					a[1][0] = 1403580;	a[1][1] = 0;	a[1][2] = 1;
					a[2][0] = -810728 + c2p32minus209;	a[2][1] = 0;	a[2][2] = 0;
					for (int i = 0; i < j; i++){
						for (int k = 0; k < 3; k++)
							for ( int h = 0; h < 3; h++){
								__int64 v = 0;
								for (int t = 0; t < 3; t++){
									__int64 vv = MultDivModP(a[k][t], a[t][h], c2p32minus209);
									v = DivModP(v + vv, c2p32minus209);
								}
								b[k][h] = v;
							}
						for (int k = 0; k < 3; k++)
							for (int h = 0; h < 3; h++)
								a[k][h] = b[k][h];
					}

					__int64 tx0 = x1_0, tx1 = x1_1, tx2 = x1_2;
					x1_0 = MultDivModP(tx0, a[0][0], c2p32minus209);
					x1_0 = DivModP( MultDivModP(tx1, a[1][0], c2p32minus209) + x1_0, c2p32minus209 );
					x1_0 = DivModP( MultDivModP(tx2, a[2][0], c2p32minus209) + x1_0, c2p32minus209 );

					x1_1 = MultDivModP(tx0, a[0][1], c2p32minus209);
					x1_1 = DivModP( MultDivModP(tx1, a[1][1], c2p32minus209) + x1_1, c2p32minus209 );
					x1_1 = DivModP( MultDivModP(tx2, a[2][1], c2p32minus209) + x1_1, c2p32minus209 );

					x1_2 = MultDivModP(tx0, a[0][2], c2p32minus209);
					x1_2 = DivModP( MultDivModP(tx1, a[1][2], c2p32minus209) + x1_2, c2p32minus209 );
					x1_2 = DivModP( MultDivModP(tx2, a[2][2], c2p32minus209) + x1_2, c2p32minus209 );

					a[0][0] = 527612;		a[0][1] = 1;	a[0][2] = 0;
					a[1][0] = 0;			a[1][1] = 0;	a[1][2] = 1;
					a[2][0] = -1370589 + c2p32minus22853;		a[2][1] = 0;	a[2][2] = 0;
					for (int i = 0; i < j; i++){
						for (int k = 0; k < 3; k++)
							for ( int h = 0; h < 3; h++){
								__int64 v = 0;
								for (int t = 0; t < 3; t++){
									__int64 vv = MultDivModP(a[k][t], a[t][h], c2p32minus22853);
									v = DivModP(v + vv, c2p32minus22853);
								}
								b[k][h] = v;
							}
						for (int k = 0; k < 3; k++)
							for (int h = 0; h < 3; h++)
								a[k][h] = b[k][h];
					}
			
					tx0 = x2_0;
					tx1 = x2_1;
					tx2 = x2_2;
					x2_0 = MultDivModP(tx0, a[0][0], c2p32minus22853);
					x2_0 = DivModP( MultDivModP(tx1, a[1][0], c2p32minus22853) + x2_0, c2p32minus22853 );
					x2_0 = DivModP( MultDivModP(tx2, a[2][0], c2p32minus22853) + x2_0, c2p32minus22853 );

					x2_1 = MultDivModP(tx0, a[0][1], c2p32minus22853);
					x2_1 = DivModP( MultDivModP(tx1, a[1][1], c2p32minus22853) + x2_1, c2p32minus22853 );
					x2_1 = DivModP( MultDivModP(tx2, a[2][1], c2p32minus22853) + x2_1, c2p32minus22853 );

					x2_2 = MultDivModP(tx0, a[0][2], c2p32minus22853);
					x2_2 = DivModP( MultDivModP(tx1, a[1][2], c2p32minus22853) + x2_2, c2p32minus22853 );
					x2_2 = DivModP( MultDivModP(tx2, a[2][2], c2p32minus22853) + x2_2, c2p32minus22853 );

					return;
				}

				
    //        Standard Normal Distribution
	//        Rejection polar method for normal variate
	double normal();


    //    Normal Distribution of mean m  and standard deviation s
	double normalMS(double m, double s);


    //    A Pair of Independent Random Numbers of 
	//      the Standard Normal Distribution 
	void  normalPair( double & n1, double & n2 );

	//		A Pair of Random Numbers of
	//		correlation coefficient r
	void normalPair_r( double & n1, double & n2, double r );

	//
	//			J. E. Gentle (2003)
	//			"Random number generation and Monte Carlo methods, 2nd ed."
	//			Springer, Chapter 5.
	//
	double gamma( double a, double b );

	//
	//			N. L. Johnson, S. Kotz and N. Balakrishnan (1995)
	//			"Continuous univariate distributions, Vol. 2, 2nd ed.
	//			John Wiley & Sons, Inc., Chapter 25
	//
	double beta( double a, double b);

	};

}
#endif
